@@ -12,6 +12,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import collections
 import contextlib
 import errno
 import functools
@@ -45,6 +46,13 @@ from oslo_messaging._drivers import pool
 from oslo_messaging import _utils
 from oslo_messaging import exceptions
 
+
+# The QuorumMemConfig will hold the quorum queue memory configurations
+QuorumMemConfig = collections.namedtuple('QuorumMemConfig',
+                                         'delivery_limit'
+                                         ' max_memory_length'
+                                         ' max_memory_bytes')
+
 # NOTE(sileht): don't exist in py2 socket module
 TCP_USER_TIMEOUT = 18
 
@@ -74,8 +82,17 @@ rabbit_opts = [
                deprecated_name='kombu_ssl_ca_certs',
                help='SSL certification authority file '
                     '(valid only if SSL enabled).'),
+    cfg.BoolOpt('ssl_enforce_fips_mode',
+                default=False,
+                help='Global toggle for enforcing the OpenSSL FIPS mode. '
+                'This feature requires Python support. '
+                'This is available in Python 3.9 in all '
+                'environments and may have been backported to older '
+                'Python versions on select environments. If the Python '
+                'executable used does not support OpenSSL FIPS mode, '
+                'an exception will be raised.'),
     cfg.BoolOpt('heartbeat_in_pthread',
-                default=True,
+                default=False,
                 help="Run the health check heartbeat thread "
                      "through a native python thread by default. If this "
                      "option is equal to False then the health check "
@@ -83,7 +100,9 @@ rabbit_opts = [
                      "from the parent process. For "
                      "example if the parent process has monkey patched the "
                      "stdlib by using eventlet/greenlet then the heartbeat "
-                     "will be run through a green thread.",
+                     "will be run through a green thread. "
+                     "This option should be set to True only for the "
+                     "wsgi services.",
                 ),
     cfg.FloatOpt('kombu_reconnect_delay',
                  default=1.0,
@@ -108,7 +127,7 @@ rabbit_opts = [
                     'unavailable. Takes effect only if more than one '
                     'RabbitMQ node is provided in config.'),
     cfg.StrOpt('rabbit_login_method',
-               choices=('PLAIN', 'AMQPLAIN', 'RABBIT-CR-DEMO'),
+               choices=('PLAIN', 'AMQPLAIN', 'EXTERNAL', 'RABBIT-CR-DEMO'),
                default='AMQPLAIN',
                deprecated_group='DEFAULT',
                help='The RabbitMQ login method.'),
@@ -147,6 +166,30 @@ rabbit_opts = [
                 'in other words the HA queues should be disabled, quorum '
                 'queues durable by default so the amqp_durable_queues '
                 'opion is ignored when this option enabled.'),
+    cfg.IntOpt('rabbit_quorum_delivery_limit',
+               default=0,
+               help='Each time a message is redelivered to a consumer, '
+               'a counter is incremented. Once the redelivery count '
+               'exceeds the delivery limit the message gets dropped '
+               'or dead-lettered (if a DLX exchange has been configured) '
+               'Used only when rabbit_quorum_queue is enabled, '
+               'Default 0 which means dont set a limit.'),
+    cfg.IntOpt('rabbit_quroum_max_memory_length',
+               default=0,
+               help='By default all messages are maintained in memory '
+               'if a quorum queue grows in length it can put memory '
+               'pressure on a cluster. This option can limit the number '
+               'of messages in the quorum queue. '
+               'Used only when rabbit_quorum_queue is enabled, '
+               'Default 0 which means dont set a limit.'),
+    cfg.IntOpt('rabbit_quroum_max_memory_bytes',
+               default=0,
+               help='By default all messages are maintained in memory '
+               'if a quorum queue grows in length it can put memory '
+               'pressure on a cluster. This option can limit the number '
+               'of memory bytes used by the quorum queue. '
+               'Used only when rabbit_quorum_queue is enabled, '
+               'Default 0 which means dont set a limit.'),
     cfg.IntOpt('rabbit_transient_queues_ttl',
                min=1,
                default=1800,
@@ -190,7 +233,8 @@ LOG = logging.getLogger(__name__)
 
 
 def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
-                         rabbit_quorum_queue):
+                         rabbit_quorum_queue,
+                         rabbit_quorum_queue_config):
     """Construct the arguments for declaring a queue.
 
     If the rabbit_ha_queues option is set, we try to declare a mirrored queue
@@ -224,6 +268,14 @@ def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
     RabbitMQ 3.8.0. If set this option will conflict with
     the HA queues (``rabbit_ha_queues``) aka mirrored queues,
     in other words HA queues should be disabled.
+
+    rabbit_quorum_queue_config:
+    Quorum queues provides three options to handle message poisoning
+    and limit the resources the qourum queue can use
+    x-delivery-limit number of times the queue will try to deliver
+    a message before it decide to discard it
+    x-max-in-memory-length, x-max-in-memory-bytes control the size
+    of memory used by quorum queue
     """
     args = {}
 
@@ -237,6 +289,15 @@ def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
 
     if rabbit_quorum_queue:
         args['x-queue-type'] = 'quorum'
+        if rabbit_quorum_queue_config.delivery_limit:
+            args['x-delivery-limit'] = \
+                rabbit_quorum_queue_config.delivery_limit
+        if rabbit_quorum_queue_config.max_memory_length:
+            args['x-max-in-memory-length'] = \
+                rabbit_quorum_queue_config.max_memory_length
+        if rabbit_quorum_queue_config.max_memory_bytes:
+            args['x-max-in-memory-bytes'] = \
+                rabbit_quorum_queue_config.max_memory_bytes
 
     if rabbit_queue_ttl > 0:
         args['x-expires'] = rabbit_queue_ttl * 1000
@@ -266,7 +327,8 @@ class Consumer(object):
     def __init__(self, exchange_name, queue_name, routing_key, type, durable,
                  exchange_auto_delete, queue_auto_delete, callback,
                  nowait=False, rabbit_ha_queues=None, rabbit_queue_ttl=0,
-                 enable_cancel_on_failover=False, rabbit_quorum_queue=False):
+                 enable_cancel_on_failover=False, rabbit_quorum_queue=False,
+                 rabbit_quorum_queue_config=QuorumMemConfig(0, 0, 0)):
         """Init the Consumer class with the exchange_name, routing_key,
         type, durable auto_delete
         """
@@ -279,18 +341,52 @@ class Consumer(object):
         self.callback = callback
         self.type = type
         self.nowait = nowait
-        self.queue_arguments = _get_queue_arguments(rabbit_ha_queues,
-                                                    rabbit_queue_ttl,
-                                                    rabbit_quorum_queue)
+        rabbit_quorum_queue_config = rabbit_quorum_queue_config or {}
+        self.queue_arguments = _get_queue_arguments(
+            rabbit_ha_queues, rabbit_queue_ttl, rabbit_quorum_queue,
+            rabbit_quorum_queue_config)
         self.queue = None
         self._declared_on = None
         self.exchange = kombu.entity.Exchange(
             name=exchange_name,
-            type=type,
+            type=self.type,
             durable=self.durable,
             auto_delete=self.exchange_auto_delete)
         self.enable_cancel_on_failover = enable_cancel_on_failover
 
+    def _declare_fallback(self, err, conn, consumer_arguments):
+        """Fallback by declaring a non durable queue.
+
+        When a control exchange is shared between services it is possible
+        that some service created first a non durable control exchange and
+        then after that an other service can try to create the same control
+        exchange but as a durable control exchange. In this case RabbitMQ
+        will raise an exception (PreconditionFailed), and then it will stop
+        our execution and our service will fail entirly. In this case we want
+        to fallback by creating a non durable queue to match the default
+        config.
+        """
+        if "PRECONDITION_FAILED - inequivalent arg 'durable'" in str(err):
+            LOG.info(
+                "[%s] Retrying to declare the exchange (%s) as "
+                "non durable", conn.connection_id, self.exchange_name)
+            self.exchange = kombu.entity.Exchange(
+                name=self.exchange_name,
+                type=self.type,
+                durable=False,
+                auto_delete=self.queue_auto_delete)
+            self.queue = kombu.entity.Queue(
+                name=self.queue_name,
+                channel=conn.channel,
+                exchange=self.exchange,
+                durable=False,
+                auto_delete=self.queue_auto_delete,
+                routing_key=self.routing_key,
+                queue_arguments=self.queue_arguments,
+                consumer_arguments=consumer_arguments
+            )
+            self.queue.declare()
+
     def declare(self, conn):
         """Re-declare the queue after a rabbit (re)connect."""
 
@@ -313,7 +409,18 @@ class Consumer(object):
         try:
             LOG.debug('[%s] Queue.declare: %s',
                       conn.connection_id, self.queue_name)
-            self.queue.declare()
+            try:
+                self.queue.declare()
+            except amqp_exec.PreconditionFailed as err:
+                # NOTE(hberaud): This kind of exception may be triggered
+                # when a control exchange is shared between services and
+                # when services try to create it with configs that differ
+                # from each others. RabbitMQ will reject the services
+                # that try to create it with a configuration that differ
+                # from the one used first.
+                LOG.warning(err)
+                self._declare_fallback(err, conn, consumer_arguments)
+
         except conn.connection.channel_errors as exc:
             # NOTE(jrosenboom): This exception may be triggered by a race
             # condition. Simply retrying will solve the error most of the time
@@ -496,6 +603,8 @@ class Connection(object):
         self.login_method = driver_conf.rabbit_login_method
         self.rabbit_ha_queues = driver_conf.rabbit_ha_queues
         self.rabbit_quorum_queue = driver_conf.rabbit_quorum_queue
+        self.rabbit_quorum_queue_config = self._get_quorum_configurations(
+            driver_conf)
         self.rabbit_transient_queues_ttl = \
             driver_conf.rabbit_transient_queues_ttl
         self.rabbit_qos_prefetch_count = driver_conf.rabbit_qos_prefetch_count
@@ -511,6 +620,7 @@ class Connection(object):
         self.kombu_failover_strategy = driver_conf.kombu_failover_strategy
         self.kombu_compression = driver_conf.kombu_compression
         self.heartbeat_in_pthread = driver_conf.heartbeat_in_pthread
+        self.ssl_enforce_fips_mode = driver_conf.ssl_enforce_fips_mode
         self.enable_cancel_on_failover = driver_conf.enable_cancel_on_failover
 
         if self.heartbeat_in_pthread:
@@ -548,6 +658,19 @@ class Connection(object):
             self.ssl_cert_file = driver_conf.ssl_cert_file
             self.ssl_ca_file = driver_conf.ssl_ca_file
 
+            if self.ssl_enforce_fips_mode:
+                if hasattr(ssl, 'FIPS_mode'):
+                    LOG.info("Enforcing the use of the OpenSSL FIPS mode")
+                    ssl.FIPS_mode_set(1)
+                else:
+                    raise exceptions.ConfigurationError(
+                        "OpenSSL FIPS mode is not supported by your Python "
+                        "version. You must either change the Python "
+                        "executable used to a version with FIPS mode "
+                        "support or disable FIPS mode by setting the "
+                        "'[oslo_messaging_rabbit] ssl_enforce_fips_mode' "
+                        "configuration option to 'False'.")
+
         self._url = ''
         if url.hosts:
             if url.transport.startswith('kombu+'):
@@ -709,6 +832,14 @@ class Connection(object):
         except KeyError:
             raise RuntimeError("Invalid SSL version : %s" % version)
 
+    def _get_quorum_configurations(self, driver_conf):
+        """Get the quorum queue configurations"""
+        delivery_limit = driver_conf.rabbit_quorum_delivery_limit
+        max_memory_length = driver_conf.rabbit_quroum_max_memory_length
+        max_memory_bytes = driver_conf.rabbit_quroum_max_memory_bytes
+        return QuorumMemConfig(delivery_limit, max_memory_length,
+                               max_memory_bytes)
+
     # NOTE(moguimar): default_password in this function's context is just
     #                 a fallback option, not a hardcoded password.
     def _transform_transport_url(self, url, host, default_username='',  # nosec
@@ -1202,7 +1333,8 @@ class Connection(object):
             callback=callback,
             rabbit_ha_queues=self.rabbit_ha_queues,
             enable_cancel_on_failover=self.enable_cancel_on_failover,
-            rabbit_quorum_queue=self.rabbit_quorum_queue)
+            rabbit_quorum_queue=self.rabbit_quorum_queue,
+            rabbit_quorum_queue_config=self.rabbit_quorum_queue_config)
 
         self.declare_consumer(consumer)
 
@@ -1266,7 +1398,20 @@ class Connection(object):
         """Publish a message."""
 
         if not (exchange.passive or exchange.name in self._declared_exchanges):
-            exchange(self.channel).declare()
+            try:
+                exchange(self.channel).declare()
+            except amqp_exec.PreconditionFailed as err:
+                # NOTE(hberaud): This kind of exception may be triggered
+                # when a control exchange is shared between services and
+                # when services try to create it with configs that differ
+                # from each others. RabbitMQ will reject the services
+                # that try to create it with a configuration that differ
+                # from the one used first.
+                if "PRECONDITION_FAILED - inequivalent arg 'durable'" \
+                   in str(err):
+                    LOG.warning("Force creating a non durable exchange.")
+                    exchange.durable = False
+                    exchange(self.channel).declare()
             self._declared_exchanges.add(exchange.name)
 
         log_info = {'msg': msg,
@@ -1317,7 +1462,8 @@ class Connection(object):
                 queue_arguments=_get_queue_arguments(
                     self.rabbit_ha_queues,
                     0,
-                    self.rabbit_quorum_queue))
+                    self.rabbit_quorum_queue,
+                    self.rabbit_quorum_queue_config))
             log_info = {'key': routing_key, 'exchange': exchange}
             LOG.trace(
                 'Connection._publish_and_creates_default_queue: '
